package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.ProtobufsMessages.BasicMessage;
import meerkat.ProtobufsMessages.Tag;
import meerkat.bulletinboard.SimpleBulletinBoardClient;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.CollectionMessagesUtils;
import util.Tags;

import java.io.IOException;
import java.text.ParseException;
import java.util.*;

/**
 *  Created by Vladimir Eliezer Tokarev on 1/8/2016.
 *  Gives the ability to manage voters information
 *  This object is synchronous (which meaning its method blocking)
 */
public class SimpleRegistry {

    protected Encryption signatory;

    protected SimpleBulletinBoardClient communicator;

    /**
     * @param signatory implements the DigitalSignature interface
     * @param communicator implements the BulletinBoardClient interface
     */
    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator){
        this.signatory = signatory;
        this.communicator = communicator;
    }

    /**
     * Adds new voter to the bulletin-board
     * @param voterID
     * @param personalData for example residence location
     * @throws throws CommunicationException
     * @return void
     */
    public void AddVoter(String voterID, String personalData) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(Tags.ID_TAG + " " + voterID);

        Tag.Builder voterEntryTag = Tag.newBuilder();
        voterEntryTag.setContent(Tags.VOTER_ENTRY_TAG.toString());

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(Tags.ACTION_TIMESTAMP_TAG + " "
                + AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(voterEntryTag).addTag(timestampTag);
        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Adding given voter to given group
     * @param voterID
     * @param groupID
     * @throws CommunicationException
     * @return true if the adding action succeeded else return false
     */
    public void AddToGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(Tags.ID_TAG + " " + voterID);

        Tag.Builder groupIDTag = Tag.newBuilder();
        groupIDTag.setContent(Tags.GROUP_ID_TAG + " " + groupID);

        Tag.Builder actionTag = Tag.newBuilder();
        actionTag.setContent(Tags.GROUP_ACTION_TAG + " " + Tags.ADD_TO_GROUP_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(Tags.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(groupIDTag).addTag(actionTag).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Removes given voter from given group
     * @param voterID
     * @param groupID
     * @return  true if the removing action succeeded else return false
     * @throws CommunicationException
     */
    public void RemoveFromGroup(String voterID, String groupID) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(Tags.ID_TAG + " " + voterID);

        Tag.Builder groupIDTag = Tag.newBuilder();
        groupIDTag.setContent(Tags.GROUP_ID_TAG + " " + groupID);

        Tag.Builder actionTag = Tag.newBuilder();
        actionTag.setContent(Tags.GROUP_ACTION_TAG + " " + Tags.REMOVE_FROM_GROUP_TAG);

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(Tags.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage =
                BasicMessage.newBuilder().addTag(idTag).addTag(groupIDTag).addTag(actionTag).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Sets that the voter have voted
     * @param id
     * @return true if the set voted succeded else false
     * @throws CommunicationException
     */
    public void SetVoted(String id) throws CommunicationException, IOException {
        Tag.Builder idTag = Tag.newBuilder();
        idTag.setContent(Tags.ID_TAG + " " + id);

        Tag.Builder voteAction = Tag.newBuilder();
        voteAction.setContent(Tags.VOTE_ACTION_TAG.toString());

        Tag.Builder timestampTag = Tag.newBuilder();
        timestampTag.setContent(Tags.ACTION_TIMESTAMP_TAG + " "
                +AccurateTimestamp.GetCurrentTimestampString());

        BasicMessage.Builder basicMessage = BasicMessage.newBuilder().addTag(idTag).addTag(voteAction).addTag(timestampTag);

        communicator.postMessage(CreateBulletinBoardMessage(basicMessage.build()));
    }

    /**
     * Creates bulletin board message from basic message (with the signatures parts)
     * @param basicMessage  BasicMessage
     * @return BulletinBoardAPI.BulletinBoardMessage
     */
    private BulletinBoardAPI.BulletinBoardMessage CreateBulletinBoardMessage(BasicMessage  basicMessage) throws IOException {

        BulletinBoardAPI.BulletinBoardMessage.Builder bulletinBoardMessage =
                BulletinBoardAPI.BulletinBoardMessage.newBuilder();

        Crypto.RerandomizableEncryptedMessage encryptedMessage =
                signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));

        Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
        messageSignature.mergeFrom(encryptedMessage);

        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder unsignedBulletinBoardMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder();
        unsignedBulletinBoardMessage.setData(basicMessage.toByteString());

        bulletinBoardMessage.addSig(messageSignature);

        bulletinBoardMessage.setMsg(unsignedBulletinBoardMessage);

        return bulletinBoardMessage.build();
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     * @param id
     * @return  List<BulletinBoardAPI.BulletinBoardMessage>
     */
    private List<BulletinBoardAPI.BulletinBoardMessage> GetRelevantMessages(String id) {
        BulletinBoardAPI.MessageFilterList.Builder filters =
                BulletinBoardAPI.MessageFilterList.newBuilder();

        BulletinBoardAPI.MessageFilter.Builder idFilter =
                BulletinBoardAPI.MessageFilter.newBuilder().setTag(Tags.ID_TAG + " " + id)
                        .setTag(Tags.GROUP_ACTION_TAG + " " + Tags.ADD_TO_GROUP_TAG)
                        .setTag(Tags.GROUP_ACTION_TAG + " " + Tags.REMOVE_FROM_GROUP_TAG);

        filters.addFilter(idFilter);

        return communicator.readMessages(filters.build());
    }


    /**
     * Requests all the groups that the given id voter is in
     * @param id
     * @return list of groups ids (or names), if the method fails its empty
     * @throws CommunicationException
     */
    public List<String> GetGroups(String id) throws CommunicationException, InvalidProtocolBufferException {
        List<BulletinBoardAPI.BulletinBoardMessage> relevantMessages = GetRelevantMessages(id);

        List<BasicMessage> messages =
                CollectionMessagesUtils.GetBasicMessagesFromBulletinBoardMessages(relevantMessages);

        List<VoterRegistryMessage> voterRegistryMessages =
                CollectionMessagesUtils.ConvertToVoterRegistryMessages(messages);
        try{
            Map<String, VoterRegistryMessage>  groupIdToMessage =
                    CollectionMessagesUtils.GetLatestGroupsActions(voterRegistryMessages);
            return CollectionMessagesUtils.GetListOfGroupIds(groupIdToMessage);
        }
        catch (ParseException e)
        {
            // write log
            return null;
        }
    }

    /**
     * Retrieves list of strings that represents voter
     * @param id
     * @return list of strings (empty list if the lookup failed)
     * @throws CommunicationException
     */
    public List<String> GetPersonIDDetails(String id) throws CommunicationException {
        /**
         *      Retrieve all List<BulletinBoardMessage> that contains this id
         *      search for message with  the wanted personal data
         *      Create list with personal data
         *      return the list
         */
    }
}
